# -*- coding: utf-8 -*-

""" Schema loading and validation.
"""

from __future__ import unicode_literals

import base64
import json
import pkgutil
from typing import Any, Dict, Hashable, List, Sequence, Text, TextIO

from future.utils import raise_from
import jsonschema

from clkhash.field_formats import FieldSpec, spec_from_json_dict


MASTER_SCHEMA_FILE_NAMES = {1: 'v1.json'}  # type: Dict[Hashable, Text]


class SchemaError(Exception):
    """ The user-defined schema is invalid.
    """


class MasterSchemaError(Exception):
    """ Master schema missing? Corrupted? Otherwise surprising? This is
        the exception for you!
    """


class GlobalHashingProperties(object):
    """ Stores global hashing properties.

        :param k: The number of bits of the hash to set per ngram.
        :param l: The length of the resulting hash in bits. This is the
            length after XOR folding.
        :param xor_folds: The number of XOR folds to perform on the hash.
        :param hash_type: The hashing function to use. Choices are
            'doubleHash' and 'blakeHash'.
        :param hash_prevent_singularity: Ignored unless hash_type is
            'doubleHash'. Prevents bloom filter collisions in certain
            cases when True.
        :param kdf_type: The key derivation function to use. Currently,
            the only permitted value is 'HKDF'.
        :param kdf_hash: The hash function to use in key derivation. The
            options are 'SHA256' and 'SHA512'.
        :param kdf_info: The info for key derivation. See documentation
            of :class:`HKDFconfig` for details.
        :param kdf_salt: The salt for key derivation. See documentation
            of :class:`HKDFconfig` for details.
        :param kdf_key_size: The size of the derived keys in bytes.
    """
    def __init__(self, **kwargs):
        # type: (...) -> None
        """ Make a GlobalHashingProperties object from keyword
            arguments.
        """
        if 'k' in kwargs:
            self.k = kwargs['k']
        if 'l' in kwargs:
            self.l = kwargs['l']
        if 'xor_folds' in kwargs:
            self.xor_folds = kwargs['xor_folds']

        if 'hash_type' in kwargs:
            self.hash_type = kwargs['hash_type']
        if 'hash_prevent_singularity' in kwargs:
            self.hash_prevent_singularity = kwargs['hash_prevent_singularity']

        if 'kdf_type' in kwargs:
            self.kdf_type = kwargs['kdf_type']
        if 'kdf_hash' in kwargs:
            self.kdf_hash = kwargs['kdf_hash']
        if 'kdf_info' in kwargs:
            self.kdf_info = kwargs['kdf_info']
        if 'kdf_salt' in kwargs:
            self.kdf_salt = kwargs['kdf_salt']
        if 'kdf_key_size' in kwargs:
            self.kdf_key_size = kwargs['kdf_key_size']


    @classmethod
    def from_json_dict(cls, properties_dict):
        # type: (Dict[str, Any]) -> GlobalHashingProperties
        """ Make a GlobalHashingProperties object from a dictionary.

            :param properties_dict:
                The dictionary must have a `'type'` key and a `'config'`
                key. The `'config'` key must map to a dictionary
                containing a `'kdf'` key, which itself maps to a
                dictionary. That dictionary must have `'type'`, `'hash'`,
                `'keySize'`, `'salt'`, and `'type'` keys.
            :return: The resulting :class:`GlobalHashingProperties` object.
        """
        result = cls()

        result.k = properties_dict['k']
        result.l = properties_dict['l']
        result.xor_folds = properties_dict.get('xor_folds', 0)

        result.hash_type = properties_dict['hash']['type']
        result.hash_prevent_singularity = properties_dict['hash'].get(
            'prevent_singularity')

        result.kdf_type = properties_dict['kdf']['type']
        result.kdf_hash = properties_dict['kdf']['hash']
        result.kdf_info = base64.b64decode(
            properties_dict['kdf']['info'])
        result.kdf_salt = base64.b64decode(
            properties_dict['kdf']['salt'])
        result.kdf_key_size = properties_dict['kdf']['keySize']

        return result


class Schema(object):
    """ Overall schema which describes how to hash plaintext identifiers
        into clks.

        :ivar version: Version for the schema. Needed to keep behaviour
            consistent between clkhash versions for the same schema.
        :ivar hashing_globals: Configuration affecting hashing of all
            fields. For example cryptographic salt material, bloom
            filter length.
        :ivar fields: Information and configuration specific to each
            field. For example how to validate and tokenize a phone
            number.
    """
    def __init__(self,
                 version,          # type: Hashable
                 hashing_globals,  # type: GlobalHashingProperties
                 fields            # type: Sequence[FieldSpec]
                 ):
        # type: (...) -> None
        self.version = version
        self.hashing_globals = hashing_globals
        self.fields = fields

    def __repr__(self):
        return "<Schema (v{}): {} fields>".format(self.version,
                                                  len(self.fields))

    @classmethod
    def from_json_dict(cls, schema_dict, validate=True):
        # type: (Dict[str, Any], bool) -> Schema
        """ Make a Schema object from a dictionary.

            :param schema_dict: This dictionary must have a `'features'`
                key specifying the columns of the dataset. It must have
                a `'version'` key containing the master schema version
                that this schema conforms to. It must have a `'hash'`
                key with all the globals.
            :param validate: (default True) Raise an exception if the
                schema does not conform to the master schema.
            :return: The resulting :class:`Schema` object.
        """
        if validate:
            # This raises iff the schema is invalid.
            validate_schema_dict(schema_dict)

        hash_properties = GlobalHashingProperties.from_json_dict(
            schema_dict['clkConfig'])
        features = schema_dict['features']
        return cls(
            schema_dict['version'],
            hash_properties,
            list(map(spec_from_json_dict, features))
        )

    @classmethod
    def from_json_file(cls, schema_file, validate=True):
        # type: (TextIO, bool) -> Schema
        """ Load a Schema object from a json file.

            :param schema_file: A JSON file containing the schema.
            :raises SchemaError: When the schema is invalid.
            :return: The resulting :class:`Schema` object.
        """
        try:
            schema_dict = json.load(schema_file)
        except ValueError as e:  # In Python 3 we can be more specific
                                 # with json.decoder.JSONDecodeError,
                                 # but that doesn't exist in Python 2.
            msg = 'The schema is not a valid JSON file.'
            raise_from(SchemaError(msg), e)

        return cls.from_json_dict(schema_dict, validate=validate)



def get_master_schema(version):
    # type: (Hashable) -> bytes
    """ Loads the master schema of given version as bytes.

        :param version: The version of the master schema whose path we
            wish to retrieve.
        :raises SchemaError: When the schema version is unknown. This
            usually means that either (a) clkhash is out of date, or (b)
            the schema version listed is incorrect.
        :return: Bytes of the schema.
    """
    try:
        file_name = MASTER_SCHEMA_FILE_NAMES[version]
    except (TypeError, KeyError) as e:
        msg = ('Schema version {} is not supported. '
               'Consider updating clkhash.').format(version)
        raise_from(SchemaError(msg), e)

    try:
        schema_bytes = pkgutil.get_data('clkhash',
            'master-schemas/{}'.format(file_name))
    except IOError as e:  # In Python 3 we can be more specific with
                          # FileNotFoundError, but that doesn't exist in
                          # Python 2.
        msg = ('The master schema could not be found. The schema cannot be '
               'validated. Please file a bug report.')
        raise_from(MasterSchemaError(msg), e)
    
    if schema_bytes is None:
        msg = ('The master schema could not be loaded. The schema cannot be '
               'validated. Please file a bug report.')
        raise MasterSchemaError(msg)

    return schema_bytes


def validate_schema_dict(schema):
    # type: (Dict[str, Any]) -> None
    """ Validate the schema.

        This raises iff either the schema or the master schema are
        invalid. If it's successful, it returns nothing.

        :param schema: The schema to validate, as parsed by `json`.
        :raises SchemaError: When the schema is invalid.
        :raises MasterSchemaError: When the master schema is invalid.
    """
    if not isinstance(schema, dict):
        msg = ('The top level of the schema file is a {}, whereas a dict is '
               'expected.'.format(type(schema).__name__))
        raise SchemaError(msg)

    if 'version' in schema:
        version = schema['version']
    else:
        raise SchemaError('A format version is expected in the schema.')

    master_schema_bytes = get_master_schema(version)
    try:
        master_schema = json.loads(master_schema_bytes.decode('utf-8'))
    except ValueError as e:  # In Python 3 we can be more specific with
                             # json.decoder.JSONDecodeError, but that
                             # doesn't exist in Python 2.
        msg = ('The master schema is not a valid JSON file. The schema cannot '
               'be validated. Please file a bug report.')
        raise_from(MasterSchemaError(msg), e)

    try:
        jsonschema.validate(schema, master_schema)
    except jsonschema.exceptions.ValidationError as e:
        raise_from(SchemaError('The schema is not valid.'), e)
    except jsonschema.exceptions.SchemaError as e:
        msg = ('The master schema is not valid. The schema cannot be '
               'validated. Please file a bug report.')
        raise_from(MasterSchemaError(msg), e)



# -*- coding: utf-8 -*-

""" Validate data inputs.

    Validates individual entries and the overall format against the
    specified schema.
"""

from typing import Sequence

from future.utils import raise_from

from clkhash.field_formats import FieldSpec, InvalidEntryError, InvalidSchemaError


class EntryError(ValueError):
    """ An entry is invalid.
    """


class FormatError(ValueError):
    """ The format of the data is invalid.
    """


def validate_data(fields,  # type: Sequence[FieldSpec]
                  data     # type: Sequence[Sequence[str]]
                  ):
    # type: (...) -> None
    """ Validate the `data` entries according to the specification in
        `fields`.

        :param fields: The `FieldSpec` objects forming the
            specification.
        :param data: The data to validate.
        :raises EntryError: When an entry is not valid according to its
            :class:`FieldSpec`.
        :raises FormatError: When the number of entries in a row does
            not match expectation.
    """
    validators = [f.validate for f in fields]

    for row in data:
        if len(validators) != len(row):
            msg = 'Row has {} entries when {} are expected.'.format(
                len(row), len(validators))
            raise FormatError(msg)

        for entry, v in zip(row, validators):
            try:
                v(entry)
            except InvalidEntryError as e:
                raise_from(EntryError('Invalid entry.'), e)


def validate_header(fields,       # type: Sequence[FieldSpec]
                    column_names  # type: Sequence[str]
                    ):
    # type: (...) -> None
    """ Validate the `column_names` according to the specification in
        `fields`.

        :param fields: The `FieldSpec` objects forming the
            specification.
        :param data: The data to validate.
        :raises FormatError: When the number of columns or the column
            identifiers don't match the specification.
    """
    if len(fields) != len(column_names):
        msg = 'Header has {} columns when {} are expected'.format(
            len(column_names), len(fields))
        raise FormatError(msg)

    for f, column in zip(fields, column_names):
        if f.identifier != column:
            msg = "Column has identifier '{}' when '{}' is expected".format(
                column, f.identifier)
            raise FormatError(msg)



import csv
from typing import AnyStr, Callable, cast, Pattern, Sequence
import re
import sys

from mypy_extensions import Arg, DefaultNamedArg


try:
    int_from_bytes = int.from_bytes
except AttributeError:
    import codecs

    def __int_from_bytes(bytes, byteorder, signed=False):
        # type: (Sequence[int], str, bool) -> int
        """ Emulate Python 3's `int.from_bytes`.

            Kudos: https://stackoverflow.com/a/30403242 (with
            modifications)

            :param bytes: The bytes to turn into an `int`.
            :param byteorder: Either `'big'` or `'little'`.
        """
        if signed:
            raise NotImplementedError(
                "Signed integers are not currently supported in this "
                "backport.")

        if byteorder == 'big':
            pass
        elif byteorder == 'little':
            bytes = bytes[::-1]
        else:
            raise ValueError("byteorder must be either 'little' or 'big'")

        hex_str = codecs.encode(bytes, 'hex')  # type: ignore
        return int(hex_str, 16)

    # Make this cast since Python 2 doesn't have syntax for default
    # named arguments. Hence, must cast so Mypy thinks it matches the
    # original function.
    int_from_bytes = cast(Callable[[Arg(Sequence[int], 'bytes'),
                                Arg(str, 'byteorder'),
                                DefaultNamedArg(bool, 'signed')],
                               int],
                      __int_from_bytes)


def re_compile_full(pattern, flags=0):
    # type: (AnyStr, int) -> Pattern
    """ Create compiled regular expression such that it matches the
        entire string. Calling re.match on the output of this function
        is equivalent to calling re.fullmatch on its input.

        This is needed to support Python 2. (On Python 3, we would just
        call re.fullmatch.)
        Kudos: https://stackoverflow.com/a/30212799

        :param pattern: The pattern to compile.
        :param flags: Regular expression flags. Refer to Python
            documentation.

        :returns: A compiled regular expression.
    """
    # A pattern of type bytes doesn't make sense in Python 3.
    assert type(pattern) is not bytes or str is bytes

    return re.compile('(?:{})\Z'.format(pattern), flags=flags)


def _utf_8_encoder(unicode_csv_data):
    return (line.encode('utf-8') for line in unicode_csv_data)


def _p2_unicode_reader(unicode_csv_data, dialect=csv.excel, **kwargs):
    """ Encode Unicode as UTF-8 and parse as CSV.

        This is needed since Python 2's `csv` doesn't do Unicode.

        Kudos: https://docs.python.org/2/library/csv.html#examples

        :param unicode_csv_data: The Unicode stream to parse.
        :param dialect: The CSV dialect to use.
        :param kwargs: Any other parameters to pass to csv.reader.

        :returns: An iterator
    """
    # Encode temporarily as UTF-8:
    utf8_csv_data = _utf_8_encoder(unicode_csv_data)

    # Now we can parse!
    csv_reader = csv.reader(utf8_csv_data, dialect=dialect, **kwargs)

    # Decode UTF-8 back to Unicode, cell by cell:
    return ([unicode(cell, 'utf-8') for cell in row] for row in csv_reader)


unicode_reader = (_p2_unicode_reader  # Python 2 with hacky workarounds.
                  if sys.version_info < (3,0)
                  else csv.reader)  # Py3 with native Unicode support.



"""
Generate CLK from data.
"""

import concurrent.futures
import logging
import platform
import sys
import time
from typing import (Any, AnyStr, Callable, Iterable, List, Optional,
                    Sequence, TextIO, Tuple, TypeVar, Union)

from tqdm import tqdm

from clkhash.backports import unicode_reader
from clkhash.bloomfilter import stream_bloom_filters, serialize_bitarray
from clkhash.key_derivation import generate_key_lists
from clkhash.schema import Schema
from clkhash.stats import OnlineMeanVariance
from clkhash.validate_data import validate_data, validate_header


log = logging.getLogger('clkhash.clk')

CHUNK_SIZE = 1000


def hash_and_serialize_chunk(chunk_pii_data,  # type: Sequence[Sequence[str]]
                             keys,            # type: Sequence[Sequence[bytes]]
                             schema           # type: Schema
                             ):
    # type: (...) -> Tuple[List[str], Sequence[int]]
    """
    Generate Bloom filters (ie hash) from chunks of PII then serialize
    the generated Bloom filters. It also computes and outputs the Hamming weight (or popcount) -- the number of bits
    set to one -- of the generated Bloom filters.

    :param chunk_pii_data: An iterable of indexable records.
    :param schema_types: An iterable of identifier type names.
    :param keys: A tuple of two lists of secret keys used in the HMAC.
    :return: A list of serialized Bloom filters and a list of corresponding popcounts
    """
    clk_data = []
    clk_popcounts = []
    for clk in stream_bloom_filters(chunk_pii_data, keys, schema):
        clk_data.append(serialize_bitarray(clk[0]).strip())
        clk_popcounts.append(clk[2])
    return clk_data, clk_popcounts


def generate_clk_from_csv(input_f,           # type: TextIO
                          keys,              # type: Tuple[AnyStr, AnyStr]
                          schema,            # type: Schema
                          validate=True,     # type: bool
                          header=True,       # type: bool
                          progress_bar=True  # type: bool
                          ):
    # type: (...) -> List[str]
    log.info("Hashing data")

    # Read from CSV file
    reader = unicode_reader(input_f)

    if header:
        column_names = next(reader)
        if validate:
            validate_header(schema.fields, column_names)

    start_time = time.time()

    # Read the lines in CSV file and add it to PII
    pii_data = []
    for line in reader:
        if len(line) == len(schema.fields):
            pii_data.append(tuple([element.strip() for element in line]))
        else:
            raise ValueError("Line had unexpected number of elements. "
                "Expected {} but there was {}".format(
                len(schema.fields), len(line)))

    if progress_bar:
        stats = OnlineMeanVariance()
        with tqdm(desc="generating CLKs", total=len(pii_data), unit='clk', unit_scale=True,
                  postfix={'mean': stats.mean(), 'std': stats.std()}) as pbar:
            def callback(tics, clk_stats):
                stats.update(clk_stats)
                pbar.set_postfix(mean=stats.mean(), std=stats.std(), refresh=False)
                pbar.update(tics)

            results = generate_clks(pii_data,
                                    schema,
                                    keys,
                                    validate=validate,
                                    callback=callback)
    else:
        results = generate_clks(pii_data,
                                schema,
                                keys,
                                validate=validate)

    log.info("Hashing took {:.2f} seconds".format(time.time() - start_time))
    return results


def generate_clks(pii_data,       # type: Sequence[Sequence[str]]
                  schema,         # type: Schema
                  keys,           # type: Tuple[AnyStr, AnyStr]
                  validate=True,  # type: bool
                  callback=None   # type: Optional[Callable[[int, Sequence[int]], None]]
                  ):
    # type: (...) -> List[str]

    # generate two keys for each identifier
    key_lists = generate_key_lists(
        keys,
        len(schema.fields),
        key_size=schema.hashing_globals.kdf_key_size,
        salt=schema.hashing_globals.kdf_salt,
        info=schema.hashing_globals.kdf_info,
        kdf=schema.hashing_globals.kdf_type,
        hash_algo=schema.hashing_globals.kdf_hash)

    if validate:
        validate_data(schema.fields, pii_data)

    # Chunks PII
    log.info("Hashing {} entities".format(len(pii_data)))
    chunk_size = 200 if len(pii_data) <= 10000 else 1000
    futures = []

    stats = OnlineMeanVariance()

    # Compute Bloom filter from the chunks and then serialise it
    with concurrent.futures.ProcessPoolExecutor() as executor:
        for chunk in chunks(pii_data, chunk_size):
            future = executor.submit(
                hash_and_serialize_chunk,
                chunk, key_lists, schema,)
            if callback is not None:
                future.add_done_callback(lambda f: callback(len(f.result()[0]), f.result()[1]))
            futures.append(future)

        results = []
        for future in futures:
            clks, clk_stats = future.result()
            stats.update(clk_stats)
            results.extend(clks)

    return results


T = TypeVar('T')      # Declare generic type variable


def chunks(seq, chunk_size):
    # type: (Sequence[T], int) -> Iterable[Sequence[T]]
    """ Split seq into chunk_size-sized chunks.

        :param seq: A sequence to chunk.
        :param chunk_size: The size of chunk.
    """
    return (seq[i:i + chunk_size] for i in range(0, len(seq), chunk_size))



"""
Module to produce a dataset of names, genders and dates of birth and manipulate that list

Currently very simple and not realistic. Additional functions for manipulating the list of names
- producing reordered and subset lists with a specific overlap

ClassList class - generate a list of length n of [id, name, dob, gender] lists

TODO: Get age distribution right by using a mortality table
TODO: Get first name distributions right by using distributions
TODO: Generate realistic errors
TODO: Add RESTfull api to generate reasonable name data as requested
"""
from __future__ import print_function

import base64
import csv
from datetime import datetime, timedelta
import math
import os
import pkgutil
import random
import re
from typing import Dict, Iterable, List, Sequence, TextIO, Tuple, Union

from clkhash.schema import Schema
from clkhash.field_formats import FieldSpec

def load_csv_data(resource_name):
    # type: (str) -> List[str]
    """Loads a specified CSV data file and returns the first column as a Python list
    """
    data_bytes = pkgutil.get_data('clkhash', 'data/{}'.format(resource_name))
    if data_bytes is None:
        raise ValueError("No data resource found with name {}".format(resource_name))
    else:
        data = data_bytes.decode('utf8')
        reader = csv.reader(data.splitlines())
        next(reader, None)  # skip the headers
        return [row[0] for row in reader]


def save_csv(data,      # type: Iterable[Tuple[Union[str, int], ...]]
             headers,   # type: Iterable[str]
             file       # type: TextIO
             ):
    # type: (...) -> None
    """
    Output generated data to file as CSV with header.

    :param data: An iterable of tuples containing raw data.
    :param schema: Iterable of schema definition dicts
    :param file: A writeable stream in which to write the CSV
    """

    print(','.join(headers), file=file)
    writer = csv.writer(file)
    writer.writerows(data)


def random_date(start, end):
    # type: (datetime, datetime) -> datetime
    """
    This function will return a random datetime between two datetime objects.

    :param start: datetime of start
    :param end: datetime of end
    :return: random datetime between start and end
    """
    delta = end - start
    int_delta = (delta.days * 24 * 60 * 60) + delta.seconds
    random_second = random.randrange(int_delta)
    return start + timedelta(seconds=random_second)


class NameList:
    """ List of randomly generated names.
    """

    with open(os.path.join(os.path.dirname(__file__),
                           'data',
                           'randomnames-schema.json')) as f:
        SCHEMA = Schema.from_json_file(f)
    del f

    def __init__(self, n):
        # type: (int) -> None
        self.load_names()

        self.earliest_birthday = datetime(year=1916, month=1, day=1)
        self.latest_birthday = datetime(year=2016, month=1, day=1)

        self.names = [person for person in self.generate_random_person(n)]

    @property
    def schema_types(self):
        # type: () -> Sequence[FieldSpec]
        return self.SCHEMA.fields

    def generate_random_person(self, n):
        # type: (int) -> Iterable[Tuple[str, str, str, str]]
        """
        Generator that yields details on a person with plausible name, sex and age.

        :yields: Generated data for one person
            tuple - (id: int, name: str('First Last'), birthdate: str('DD/MM/YYYY'), sex: str('M' | 'F') )
        """
        for i in range(n):
            sex = 'M' if random.random() > 0.5 else 'F'
            dob = random_date(self.earliest_birthday, self.latest_birthday).strftime("%Y/%m/%d")
            first_name = random.choice(self.all_male_first_names) if sex == 'M' else random.choice(self.all_female_first_names)
            last_name = random.choice(self.all_last_names)

            yield (
                str(i),
                first_name + ' ' + last_name,
                dob,
                sex
            )

    def load_names(self):
        # type: () -> None
        """
        This function loads a name database into globals firstNames and lastNames

        initial version uses data files from
        http://www.quietaffiliate.com/free-first-name-and-last-name-databases-csv-and-sql/

        """

        self.all_male_first_names = load_csv_data('male-first-names.csv')
        self.all_female_first_names = load_csv_data('female-first-names.csv')
        self.all_last_names = load_csv_data('CSV_Database_of_Last_Names.csv')

    def generate_subsets(self, sz, overlap=0.8):
        """
        Return a pair of random subsets of the name list with a specified
        proportion of elements in common.

        :param sz: length of subsets to generate
        :param overlap: fraction of the subsets that should have the same names in them
        :raises ValueError: if there aren't sufficiently many names in the list to satisfy
            the request; more precisely, raises if sz + (1 - overlap)*sz > n = len(self.names)
        :return: pair of subsets
        """
        notoverlap = sz - int(math.floor(overlap * sz))
        total_sz = sz + notoverlap
        if total_sz > len(self.names):
            raise ValueError('Requested subset size and overlap demands more '
                             + 'than the number of available names')
        sset = random.sample(self.names, total_sz)
        return sset[:sz], sset[notoverlap:]



# -*- coding: utf-8 -*-

"""
Functions to tokenize words (PII)
"""
from __future__ import unicode_literals

import functools
from typing import AnyStr, Callable, Iterable, Optional, Text

from clkhash import field_formats


def tokenize(n, positional, word, ignore=None):
    # type: (int, bool, Text, Optional[Text]) -> Iterable[Text]
    """ Produce `n`-grams of `word`.

        :param n: Length of `n`-grams.
        :param positional: If `True`, then include the index of the
            substring with the `n`-gram.
        :param word: The string to tokenize.
        :param ignore: The substring whose occurrences we remove from
            `word` before tokenization.
        :raises ValueError: When `n` is negative.
        :return: Tuple of n-gram strings.
    """
    if n < 0:
        raise ValueError('`n` in `n`-gram must be non-negative.')

    if ignore is not None:
        word = word.replace(ignore, '')

    if n > 1:
        word = ' {} '.format(word)

    if positional:
        # These are 1-indexed.
        return ('{} {}'.format(i + 1, word[i:i+n])
                for i in range(len(word) - n + 1))
    else:
        return (word[i:i+n] for i in range(len(word) - n + 1))



def get_tokenizer(hash_settings  # type: field_formats.FieldHashingProperties
                  ):
    # type: (...) -> Callable[[Text], Iterable[Text]]
    """ Get tokeniser function from the hash settings.

        This function takes a FieldHashingProperties object. It returns a
        function that takes a string and tokenises based on those properties.
    """
    n = hash_settings.ngram
    p = hash_settings.positional

    return functools.partial(tokenize, n, p)



# -*- coding: utf-8 -*-

""" Classes that specify the requirements for each column in a dataset.
    They take care of validation, and produce the settings required to
    perform the hashing.
"""

from __future__ import unicode_literals

import abc
from datetime import datetime
import re
from typing import Any, cast, Dict, Iterable, Optional, Text, Union

from future.builtins import super
from future.utils import raise_from
from six import add_metaclass

from clkhash.backports import re_compile_full


class InvalidEntryError(ValueError):
    """ An entry in the data file does not conform to the schema.
    """


class InvalidSchemaError(ValueError):
    """ The schema is not valid.

        This exception is raised if, for example, a regular expression
        included in the schema is not syntactically correct.
    """


class FieldHashingProperties(object):
    """ Stores the settings used to hash a field. This includes the
        encoding and tokenisation parameters.

        :ivar str encoding: The encoding to use when converting the
            string to bytes. Refer to
            `Python's documentation <https://docs.python.org/3/library/codecs.html#standard-encodings>`_
            for possible values.
        :ivar int ngram: The n in n-gram. Possible values are 0, 1, and
            2.
        :ivar bool positional: Controls whether the n-grams are
            positional.
        :ivar float weight: Controls the weight of the field in the
            Bloom filter.
    """
    _DEFAULT_ENCODING = 'utf-8'
    _DEFAULT_POSITIONAL = False
    _DEFAULT_WEIGHT = 1

    def __init__(self,
                 ngram,                          # type: int
                 encoding=_DEFAULT_ENCODING,     # type: str
                 weight=_DEFAULT_WEIGHT,         # type: Union[int, float]
                 positional=_DEFAULT_POSITIONAL  # type: bool
                 ):
        # type: (...) -> None
        """ Make a :class:`FieldHashingProperties` object, setting it
            attributes to values specified in keyword arguments.
        """
        if ngram not in range(3):
            msg = 'ngram is {} but is expected to be 0, 1, or 2.'
            raise ValueError(msg.format(ngram))

        try:
            ''.encode(encoding)
        except LookupError as e:
            msg = '{} is not a valid Python encoding.'
            raise_from(ValueError(msg.format(encoding)), e)

        if weight < 0:
            msg = 'weight should be non-negative but is {}.'
            raise ValueError(msg.format(weight))

        self.ngram = ngram
        self.encoding = encoding
        self.positional = positional
        self.weight = weight

    @classmethod
    def from_json_dict(cls, json_dict):
        # type: (Dict[str, Any]) -> FieldHashingProperties
        """ Make a :class:`FieldHashingProperties` object from a
            dictionary.

            :param dict json_dict:
                The dictionary must have have an 'ngram' key. It may have
                'positional' and 'weight' keys; if these are missing, then
                they are filled with the default values. The encoding is
                always set to the default value.
            :return: A :class:`FieldHashingProperties` instance.
        """
        return cls(
            ngram=json_dict['ngram'],
            positional=json_dict.get(
                'positional', FieldHashingProperties._DEFAULT_POSITIONAL),
            weight=json_dict.get(
                'weight', FieldHashingProperties._DEFAULT_WEIGHT))


@add_metaclass(abc.ABCMeta)
class FieldSpec(object):
    """ Abstract base class representing the specification of a column
        in the dataset. Subclasses validate entries, and modify the
        `hashing_properties  ivar to customise hashing procedures.

        :ivar str identifier: The name of the field.
        :ivar str description: Description of the field format.
        :ivar FieldHashingProperties hashing_properties: The properties
            for hashing.
    """
    def __init__(self,
                 identifier,          # type: str
                 hashing_properties,  # type: FieldHashingProperties
                 description=None     # type: Optional[str]
                 ):
        # type: (...) -> None
        """ Make a FieldSpec object, setting it attributes to values
            specified in keyword arguments.
        """
        self.identifier = identifier
        self.hashing_properties = hashing_properties
        self.description = description

    @classmethod
    def from_json_dict(cls, field_dict):
        # type: (Dict[str, Any]) -> FieldSpec
        """ Initialise a FieldSpec object from a dictionary of
            properties.

            :param dict field_dict: The properties dictionary to use. Must
                contain a `'hashing'` key that meets the requirements of
                :class:`FieldHashingProperties`. Subclasses may requrire
            :raises InvalidSchemaError: When the `properties`
                dictionary contains invalid values. Exactly what that
                means is decided by the subclasses.
        """
        identifier = field_dict['identifier']
        description = field_dict['format'].get('description')
        hashing_properties = FieldHashingProperties.from_json_dict(
            field_dict['hashing'])

        result = cls.__new__(cls)  # type: ignore
        result.identifier = identifier
        result.hashing_properties = hashing_properties
        result.description = description

        return result

    @abc.abstractmethod
    def validate(self, str_in):
        # type: (Text) -> None
        """ Validates an entry in the field.

            Raises :class:`InvalidEntryError` iff the entry is invalid.

            Subclasses must override this method with their own
            validation. They should call the parent's `validate` method
            via `super`.

            :param str str_in: String to validate.
            :raises InvalidEntryError: When entry is invalid.
        """
        try:
            str_in.encode(encoding=self.hashing_properties.encoding)
        except UnicodeEncodeError as e:
            msg = ("Expected entry that can be encoded in {}. Read '{}'."
                   .format(self.hashing_properties.encoding, str_in))
            raise_from(InvalidEntryError(msg), e)


class StringSpec(FieldSpec):
    """ Represents a field that holds strings.

        One way to specify the format of the entries is to provide a
        regular expression that they must conform to. Another is to
        provide zero or more of: minimum length, maximum length, casing
        (lower, upper, mixed).

        Each string field also specifies an encoding used when turning
        characters into bytes. This is stored in `hashing_properties`
        since it is needed for hashing.

        :ivar regex: Compiled regular expression that entries must
            conform to. Present only if the specification is regex-
            -based.
        :ivar str case: The casing of the entries. One of `'lower'`,
            `'upper'`, or `'mixed'`. Default is `'mixed'`. Present only
            if the specification is not regex-based.
        :ivar int min_length: The minimum length of the string. `None`
            if there is no minimum length. Present only if the
            specification is not regex-based.
        :ivar int max_length: The maximum length of the string. `None`
            if there is no maximum length. Present only if the
            specification is not regex-based.
    """
    _DEFAULT_CASE = 'mixed'
    _DEFAULT_MIN_LENGTH = 0
    _PERMITTED_CASE_STYLES = {'lower', 'upper', 'mixed'}

    def __init__(self,
                 identifier,                      # type: str
                 hashing_properties,              # type: FieldHashingProperties
                 description=None,                # type: str
                 regex=None,                      # type: Optional[str]
                 case=_DEFAULT_CASE,              # type: str
                 min_length=_DEFAULT_MIN_LENGTH,  # type: Optional[int]
                 max_length=None                  # type: Optional[int]
                 ):
        # type: (...) -> None
        """ Make a StringSpec object, setting it attributes to values
            specified in keyword arguments.
        """ 
        super().__init__(identifier=identifier,
                         description=description,
                         hashing_properties=hashing_properties)

        regex_based = regex is not None

        if regex_based and (case != self._DEFAULT_CASE
                            or min_length != self._DEFAULT_MIN_LENGTH
                            or max_length is not None):
            msg = ('regex cannot be passed along with case, min_length, or'
                   ' max_length.')
            raise ValueError(msg)

        if case not in self._PERMITTED_CASE_STYLES:
            msg = ("the case is {}, but should be 'lower', 'upper', or"
                   "'mixed'")
            raise ValueError(msg.format(case))

        if regex_based and min_length < 0:
            msg = ('min_length must be non-negative, but is {}')
            raise ValueError(msg.format(min_length))

        if regex_based and max_length is not None and max_length <= 0:
            msg = ('max_length must be positive, but is {}')
            raise ValueError(msg.format(max_length))

        if regex_based:
            try:
                compiled_regex = re_compile_full(regex)
            except (SyntaxError, re.error) as e:
                msg = "invalid regular expression '{}.'".format(regex)
                raise_from(InvalidSchemaError(msg), e)
            self.regex = compiled_regex
        else:
            self.case = case
            self.min_length = min_length
            self.max_length = max_length

        self.regex_based = regex_based

    @classmethod
    def from_json_dict(cls, json_dict):
        # type: (Dict[str, Any]) -> StringSpec
        """ Make a StringSpec object from a dictionary containing its
            properties.

            :param dict json_dict: This dictionary must contain an
                `'encoding'` key associated with a Python-conformant
                encoding. It must also contain a `'hashing'` key, whose
                contents are passed to :class:`FieldHashingProperties`.
                Permitted keys also include `'pattern'`, `'case'`,
                `'minLength'`, and `'maxLength'`.
            :raises InvalidSchemaError: When a regular expression is
                provided but is not a valid pattern.
        """
        result = cast(StringSpec,  # Go away, Mypy.
                      super().from_json_dict(json_dict))

        format_ = json_dict['format']
        result.hashing_properties.encoding = format_['encoding']

        if 'pattern' in format_:
            pattern = format_['pattern']
            try:
                result.regex = re_compile_full(pattern)
            except (SyntaxError, re.error) as e:
                msg = "Invalid regular expression '{}.'".format(pattern)
                raise_from(InvalidSchemaError(msg), e)
            result.regex_based = True

        else:
            result.case = format_.get('case', StringSpec._DEFAULT_CASE)
            result.min_length = format_.get('minLength')
            result.max_length = format_.get('maxLength')
            result.regex_based = False

        return result

    def validate(self, str_in):
        # type: (Text) -> None
        """ Validates an entry in the field.

            Raises `InvalidEntryError` iff the entry is invalid.

            An entry is invalid iff (1) a pattern is part of the
            specification of the field and the string does not match
            it; (2) the string does not match the provided casing,
            minimum length, or maximum length; or (3) the specified
            encoding cannot represent the string.

            :param str str_in: String to validate.
            :raises InvalidEntryError: When entry is invalid.
            :raises ValueError: When self.case is not one of the
                permitted values (`'lower'`, `'upper'`, or `'mixed'`).
        """
        super().validate(str_in)  # Validate encoding.

        if self.regex_based:
            match = self.regex.match(str_in)
            if match is None:
                raise InvalidEntryError(
                    'Expected entry that conforms to regular expression '
                    "'{}'. Read '{}'.".format(self.regex.pattern, str_in))

        else:
            str_len = len(str_in)
            if self.min_length is not None and str_len < self.min_length:
                raise InvalidEntryError(
                    'Expected string length of at least {}. Read string of '
                    'length {}.'.format(self.min_length, str_len))

            if self.max_length is not None and str_len > self.max_length:
                raise InvalidEntryError(
                    'Expected string length of at most {}. Read string of '
                    'length {}.'.format(self.max_length, str_len))

            if self.case == 'upper':
                if str_in.upper() != str_in:
                    raise InvalidEntryError(
                        'Expected upper case string. Read {}.'.format(str_in))
            elif self.case == 'lower':
                if str_in.lower() != str_in:
                    raise InvalidEntryError(
                        'Expected lower case string. Read {}.'.format(str_in))
            elif self.case == 'mixed':
                pass
            else:
                raise ValueError(
                    'Invalid case property {}.'.format(self.case))


class IntegerSpec(FieldSpec):
    """ Represents a field that holds integers.

        Minimum and maximum values may be specified.

        :ivar int minimum: The minimum permitted value.
        :ivar int maximum: The maximum permitted value or None.
    """

    _DEFAULT_MINIMUM = 0

    def __init__(self,
                 identifier,                # type: str
                 hashing_properties,        # type: FieldHashingProperties
                 description=None,          # type: str
                 minimum=_DEFAULT_MINIMUM,  # int
                 maximum=None,              # Optional[int]
                 **kwargs                   # Dict[str, Any]
                 ):
        # type: (...) -> None
        """ Make a IntegerSpec object, setting it attributes to values
            specified in keyword arguments.
        """
        super().__init__(identifier=identifier,
                         description=description,
                         hashing_properties=hashing_properties)

        self.minimum = minimum
        self.maximum = maximum

    @classmethod
    def from_json_dict(cls, json_dict):
        # type: (Dict[str, Any]) -> IntegerSpec
        """ Make a IntegerSpec object from a dictionary containing its
            properties.

            :param dict json_dict: This dictionary may contain
                `'minimum'` and `'maximum'` keys. In addition, it must
                contain a `'hashing'` key, whose contents are passed to
                :class:`FieldHashingProperties`.

            :param dict json_dict: The properties dictionary.
        """
        result = cast(IntegerSpec,  # For Mypy.
                      super().from_json_dict(json_dict))

        format_ = json_dict['format']
        result.minimum = format_.get('minimum', cls._DEFAULT_MINIMUM)
        result.maximum = format_.get('maximum')

        return result

    def validate(self, str_in):
        # type: (Text) -> None
        """ Validates an entry in the field.

            Raises `InvalidEntryError` iff the entry is invalid.

            An entry is invalid iff (1) the string does not represent a
            base-10 integer; (2) the integer is not between
            `self.minimum` and `self.maximum`, if those exist; or (3)
            the integer is negative.

            :param str str_in: String to validate.
            :raises InvalidEntryError: When entry is invalid.
        """
        super().validate(str_in)

        try:
            value = int(str_in, base=10)
        except ValueError as e:
            msg = 'Invalid integer. Read {}.'.format(str_in)
            raise_from(InvalidEntryError(msg), e)

        if value < self.minimum:
            msg = ('Expected integer value of at least {}. Read {}.'
                   .format(self.minimum, value))
            raise InvalidEntryError(msg)

        if self.maximum is not None and value > self.maximum:
            msg = ('Expected integer value of at most {}. Read {}.'
                   .format(self.maximum, value))
            raise InvalidEntryError(msg)


class DateSpec(FieldSpec):
    """ Represents a field that holds dates.

       Dates are specified as full-dates as defined in
       `RFC3339 <https://tools.ietf.org/html/rfc3339>`_ E.g.,
       ``1996-12-19``

        :ivar str format: The format of the date.
    """
    _PERMITTED_FORMATS = {'rfc3339'}
    _RFC3339_REGEX = re_compile_full(r'\d\d\d\d-\d\d-\d\d')
    _RFC3339_FORMAT = '%Y-%m-%d'

    def __init__(self,
                 identifier,          # type: str
                 hashing_properties,  # type: FieldHashingProperties
                 format,              # type: str
                 description=None     # type: str
                 ):
        # type: (...) -> None
        """ Make a DateSpec object, setting it attributes to values
            specified in keyword arguments.
        """
        super().__init__(identifier=identifier,
                         description=description,
                         hashing_properties=hashing_properties)

        if format not in self._PERMITTED_FORMATS:
            msg = 'No validation for date format: {}.'.format(format)
            raise NotImplementedError(msg)

        self.format = format

    @classmethod
    def from_json_dict(cls, json_dict):
        # type: (Dict[str, Any]) -> DateSpec
        """ Make a DateSpec object from a dictionary containing its
            properties.

            :param dict json_dict: This dictionary must contain a
                `'format'` key. In addition, it must contain a
                `'hashing'` key, whose contents are passed to
                :class:`FieldHashingProperties`.

            :param json_dict: The properties dictionary.
        """
        result = cast(DateSpec,  # For Mypy.
                      super().from_json_dict(json_dict))

        format_ = json_dict['format']
        result.format = format_['format']

        return result

    def validate(self, str_in):
        # type: (Text) -> None
        """ Validates an entry in the field.

            Raises `InvalidEntryError` iff the entry is invalid.

            An entry is invalid iff (1) the string does not represent a
            date in the correct format; or (2) the date it represents
            is invalid (such as 30 February).

            :param str str_in: String to validate.
            :raises InvalidEntryError: Iff entry is invalid.
            :raises ValueError: When self.format is unrecognised.
        """
        super().validate(str_in)

        if self.format == 'rfc3339':
            if self._RFC3339_REGEX.match(str_in) is None:
                msg = ('Date expected to conform to RFC3339. Read {}.'
                       .format(str_in))
                raise InvalidEntryError(msg)
            try:
                datetime.strptime(str_in, self._RFC3339_FORMAT)
            except ValueError as e:
                msg = 'Invalid date. Read {}.'.format(str_in)
                raise_from(InvalidEntryError(msg), e)

        else:
            msg = 'No validation for date format: {}.'.format(self.format)
            raise NotImplementedError(msg)


class EnumSpec(FieldSpec):
    """ Represents a field that holds an enum.

        The finite collection of permitted values must be specified.

        :ivar values: The set of permitted values.
    """
    def __init__(self,
                 identifier,          # type: str
                 hashing_properties,  # type: FieldHashingProperties
                 values,              # type: Iterable[str]
                 description=None     # type: str
                 ):
        # type: (...) -> None
        """ Make a EnumSpec object, setting it attributes to values
            specified in keyword arguments.
        """
        super().__init__(identifier=identifier,
                         description=description,
                         hashing_properties=hashing_properties)

        self.values = set(values)

    @classmethod
    def from_json_dict(cls, json_dict):
        # type: (Dict[str, Any]) -> EnumSpec
        """ Make a EnumSpec object from a dictionary containing its
            properties.

            :param dict json_dict: This dictionary must contain an
                `'enum'` key specifying the permitted values. In
                addition, it must contain a `'hashing'` key, whose
                contents are passed to :class:`FieldHashingProperties`.
        """
        result = cast(EnumSpec,  # Appease the gods of Mypy.
                      super().from_json_dict(json_dict))

        format_ = json_dict['format']
        result.values = set(format_['values'])

        return result

    def validate(self, str_in):
        # type: (Text) -> None
        """ Validates an entry in the field.

            Raises `InvalidEntryError` iff the entry is invalid.

            An entry is invalid iff it is not one of the permitted
            values.

            :param str str_in: String to validate.
            :raises InvalidEntryError: When entry is invalid.
        """
        super().validate(str_in)

        if str_in not in self.values:
            msg = ('Expected enum value is one of {}. Read {}.'
                   .format(self.values, str_in))
            raise InvalidEntryError(msg)


# Map type string (as defined in master schema) to
FIELD_TYPE_MAP = {
    'string': StringSpec,
    'integer': IntegerSpec,
    'date': DateSpec,
    'enum': EnumSpec,
}


def spec_from_json_dict(json_dict):
    # type: (Dict[str, Any]) -> FieldSpec
    """ Turns a dictionary into the appropriate object.

        :param dict json_dict: A dictionary with properties.
        :returns: An initialised instance of the appropriate FieldSpec
            subclass.
    """
    type_str = json_dict['format']['type']
    spec_type = cast(FieldSpec, FIELD_TYPE_MAP[type_str])
    return spec_type.from_json_dict(json_dict)
